@@ -42,12 +42,13 @@ import orca.orca_state as orca_state
 import orca.speech as speech
 import orca.settings as settings
 import orca.settings_manager as settings_manager
+import orca.structural_navigation as structural_navigation
 
-from .speech_generator import SpeechGenerator
 from .braille_generator import BrailleGenerator
 from .formatting import Formatting
-from .structural_navigation import StructuralNavigation
 from .script_utilities import Utilities
+from .spellcheck import SpellCheck
+from .speech_generator import SpeechGenerator
 
 _settingsManager = settings_manager.getManager()
 
@@ -63,6 +64,7 @@ class Script(default.Script):
         default.Script.__init__(self, app)
 
         self.speakSpreadsheetCoordinatesCheckButton = None
+        self.alwaysSpeakSelectedSpreadsheetRangeCheckButton = None
         self.skipBlankCellsCheckButton = None
         self.speakCellCoordinatesCheckButton = None
         self.speakCellHeadersCheckButton = None
@@ -77,15 +79,6 @@ class Script(default.Script):
         self.dynamicColumnHeaders = {}
         self.dynamicRowHeaders = {}
 
-        # The following variables will be used to try to determine if we've
-        # already handled this misspelt word (see readMisspeltWord() for
-        # more details.
-
-        self.lastTextLength = -1
-        self.lastBadWord = ''
-        self.lastStartOff = -1
-        self.lastEndOff = -1
-
     def getBrailleGenerator(self):
         """Returns the braille generator for this script.
         """
@@ -96,6 +89,11 @@ class Script(default.Script):
         """
         return SpeechGenerator(self)
 
+    def getSpellCheck(self):
+        """Returns the spellcheck for this script."""
+
+        return SpellCheck(self)
+
     def getFormatting(self):
         """Returns the formatting strings for this script."""
         return Formatting(self)
@@ -109,14 +107,14 @@ class Script(default.Script):
         """Returns the 'structural navigation' class for this script.
         """
         types = self.getEnabledStructuralNavigationTypes()
-        return StructuralNavigation(self, types, enabled=False)
+        return structural_navigation.StructuralNavigation(self, types, enabled=False)
 
     def getEnabledStructuralNavigationTypes(self):
         """Returns a list of the structural navigation object types
         enabled in this script.
         """
 
-        enabledTypes = [StructuralNavigation.TABLE_CELL]
+        enabledTypes = [structural_navigation.StructuralNavigation.TABLE_CELL]
 
         return enabledTypes
 
@@ -168,6 +166,11 @@ class Script(default.Script):
                 cmdnames.PAN_BRAILLE_RIGHT,
                 False) # Do not enable learn mode for this action
 
+        self.inputEventHandlers["whereAmISelectionHandler"] = \
+            input_event.InputEventHandler(
+                Script.whereAmISelection,
+                cmdnames.WHERE_AM_I_SELECTION)
+
     def getAppKeyBindings(self):
         """Returns the application-specific keybindings for this script."""
 
@@ -232,8 +235,15 @@ class Script(default.Script):
         self.speakSpreadsheetCoordinatesCheckButton.set_active(value)
         grid.attach(self.speakSpreadsheetCoordinatesCheckButton, 0, 0, 1, 1)
 
+        label = guilabels.SPREADSHEET_SPEAK_SELECTED_RANGE
+        value = _settingsManager.getSetting('alwaysSpeakSelectedSpreadsheetRange')
+        self.alwaysSpeakSelectedSpreadsheetRangeCheckButton = \
+            Gtk.CheckButton.new_with_mnemonic(label)
+        self.alwaysSpeakSelectedSpreadsheetRangeCheckButton.set_active(value)
+        grid.attach(self.alwaysSpeakSelectedSpreadsheetRangeCheckButton, 0, 1, 1, 1)
+
         tableFrame = Gtk.Frame()
-        grid.attach(tableFrame, 0, 1, 1, 1)
+        grid.attach(tableFrame, 0, 2, 1, 1)
 
         label = Gtk.Label(label="<b>%s</b>" % guilabels.TABLE_NAVIGATION)
         label.set_use_markup(True)
@@ -273,6 +283,8 @@ class Script(default.Script):
         self.skipBlankCellsCheckButton.set_active(value)
         tableGrid.attach(self.skipBlankCellsCheckButton, 0, 3, 1, 1)
 
+        spellcheck = self.spellcheck.getAppPreferencesGUI()
+        grid.attach(spellcheck, 0, len(grid.get_children()), 1, 1)
         grid.show_all()
 
         return grid
@@ -280,34 +292,25 @@ class Script(default.Script):
     def getPreferencesFromGUI(self):
         """Returns a dictionary with the app-specific preferences."""
 
-        return {
+        prefs = {
             'speakCellSpan': self.speakCellSpanCheckButton.get_active(),
             'speakCellHeaders': self.speakCellHeadersCheckButton.get_active(),
             'skipBlankCells': self.skipBlankCellsCheckButton.get_active(),
             'speakCellCoordinates': self.speakCellCoordinatesCheckButton.get_active(),
             'speakSpreadsheetCoordinates': self.speakSpreadsheetCoordinatesCheckButton.get_active(),
+            'alwaysSpeakSelectedSpreadsheetRange': self.alwaysSpeakSelectedSpreadsheetRangeCheckButton.get_active(),
         }
 
-    def isStructuralNavigationCommand(self, inputEvent=None):
-        """Checks to see if the inputEvent was a structural navigation
-        command. This is necessary to prevent double-presentation of
-        items. [[[TODO - JD: Perhaps this should be moved to default.py]]]
+        prefs.update(self.spellcheck.getPreferencesFromGUI())
+        return prefs
 
-        Arguments:
-        - inputEvent: The input event to examine. If none is provided,
-          orca_state.lastInputEvent will be used.
+    def presentObject(self, obj, **args):
+        if not self._lastCommandWasStructNav:
+            super().presentObject(obj, **args)
+            return
 
-        Returns True if inputEvent is a structural navigation command
-        enabled in this script.
-        """
-
-        inputEvent = inputEvent or orca_state.lastInputEvent
-        if isinstance(inputEvent, input_event.KeyboardEvent):
-            if self.structuralNavigation.keyBindings.\
-                    getInputHandler(inputEvent):
-                return True
-
-        return False
+        utterances = self.speechGenerator.generateSpeech(obj, **args)
+        speech.speak(utterances)
 
     def panBrailleLeft(self, inputEvent=None, panAmount=0):
         """In document content, we want to use the panning keys to browse the
@@ -431,35 +434,13 @@ class Script(default.Script):
         row, column, table = self.utilities.getRowColumnAndTable(cell)
         try:
             del self.dynamicColumnHeaders[hash(table)]
-            speech.stop()
+            self.presentationInterrupt()
             self.presentMessage(messages.DYNAMIC_COLUMN_HEADER_CLEARED)
         except:
             pass
 
         return True
 
-    def columnConvert(self, column):
-        """ Convert a spreadsheet column into it's column label
-
-        Arguments:
-        - column: the column number to convert.
-
-        Returns a string representing the spread sheet column.
-        """
-
-        base26 = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
-
-        if column <= len(base26):
-            return base26[column-1]
-
-        res = ""
-        while column > 0:
-            digit = column % len(base26)
-            res = " " + base26[digit-1] + res
-            column /= len(base26)
-
-        return res
-
     def setDynamicRowHeaders(self, inputEvent):
         """Set the column for the dynamic header rows to use when speaking
         calc cell entries. In order to set the column, the user should first
@@ -481,7 +462,7 @@ class Script(default.Script):
         if table:
             self.dynamicRowHeaders[hash(table)] = column
             self.presentMessage(
-                messages.DYNAMIC_ROW_HEADER_SET % self.columnConvert(column+1))
+                messages.DYNAMIC_ROW_HEADER_SET % self.utilities.columnConvert(column+1))
 
         return True
 
@@ -499,114 +480,13 @@ class Script(default.Script):
         row, column, table = self.utilities.getRowColumnAndTable(cell)
         try:
             del self.dynamicRowHeaders[hash(table)]
-            speech.stop()
+            self.presentationInterrupt()
             self.presentMessage(messages.DYNAMIC_ROW_HEADER_CLEARED)
         except:
             pass
 
         return True
 
-    def readMisspeltWord(self, event, pane):
-        """Speak/braille the current misspelt word plus its context.
-           The spell check dialog contains a "paragraph" which shows the
-           context for the current spelling mistake. After speaking/brailling
-           the default action for this component, that a selection of the
-           surronding text from that paragraph with the misspelt word is also
-           spoken.
-
-        Arguments:
-        - event: the event.
-        - pane: the option pane in the spell check dialog.
-
-        Returns True if this is the spell check dialog (whether we actually
-        wind up reading the word or not).
-        """
-
-        paragraph = self.utilities.descendantsWithRole(
-            pane, pyatspi.ROLE_PARAGRAPH)
-
-        # If there is not exactly one paragraph, this isn't the spellcheck
-        # dialog.
-        #
-        if len(paragraph) != 1:
-            return False
-
-        # If there's not any text displayed in the paragraph, this isn't
-        # the spellcheck dialog.
-        #
-        try:
-            text = paragraph[0].queryText()
-        except:
-            return False
-        else:
-            textLength = text.characterCount
-            if not textLength:
-                return False
-
-        # If the text here is not editable, this isn't the spellcheck
-        # dialog.
-        #
-        if not paragraph[0].getState().contains(pyatspi.STATE_EDITABLE):
-            return False
-
-        # Determine which word is the misspelt word. This word will have
-        # non-default text attributes associated with it.
-        #
-        startFound = False
-        startOff = 0
-        endOff = textLength
-        for i in range(0, textLength):
-            attributes = text.getAttributes(i)
-            if len(attributes[0]) != 0:
-                if not startFound:
-                    startOff = i
-                    startFound = True
-            else:
-                if startFound:
-                    endOff = i
-                    break
-
-        if not startFound:
-            # If there are no text attributes in this paragraph, this isn't
-            # the spellcheck dialog.
-            #
-            return False
-
-        badWord = self.utilities.substring(paragraph[0], startOff, endOff - 1)
-
-        # Note that we often get two or more of these focus or property-change
-        # events each time there is a new misspelt word. We extract the
-        # length of the line of text, the misspelt word, the start and end
-        # offsets for that word and compare them against the values saved
-        # from the last time this routine was called. If they are the same
-        # then we ignore it.
-        #
-        debug.println(debug.LEVEL_INFO,
-            "StarOffice.readMisspeltWord: type=%s  word=%s(%d,%d)  len=%d" % \
-            (event.type, badWord, startOff, endOff, textLength))
-
-        if (textLength == self.lastTextLength) and \
-           (badWord == self.lastBadWord) and \
-           (startOff == self.lastStartOff) and \
-           (endOff == self.lastEndOff):
-            return True
-
-        # Create a list of all the words found in the misspelt paragraph.
-        #
-        text = self.utilities.substring(paragraph[0], 0, -1)
-        allTokens = text.split()
-        self.speakMisspeltWord(allTokens, badWord)
-
-        # Save misspelt word information for comparison purposes next
-        # time around.
-        #
-        self.lastTextLength = textLength
-        self.lastBadWord = badWord
-        self.lastStartOff = startOff
-        self.lastEndOff = endOff
-
-        return True
-
     def locusOfFocusChanged(self, event, oldLocusOfFocus, newLocusOfFocus):
         """Called when the visual object with focus changes.
 
@@ -628,6 +508,13 @@ class Script(default.Script):
         if self.flatReviewContext:
             self.toggleFlatReviewMode()
 
+        if self.spellcheck.isSuggestionsItem(newLocusOfFocus) \
+           and not self.spellcheck.isSuggestionsItem(oldLocusOfFocus):
+            orca.emitRegionChanged(newFocus)
+            self.updateBraille(newLocusOfFocus)
+            self.spellcheck.presentSuggestionListItem(includeLabel=True)
+            return
+
         # TODO - JD: Sad hack that wouldn't be needed if LO were fixed.
         # If we are in the slide presentation scroll pane, also announce
         # the current page tab. See bug #538056 for more details.
@@ -680,36 +567,19 @@ class Script(default.Script):
         # Pass the event onto the parent class to be handled in the default way.
         default.Script.locusOfFocusChanged(self, event,
                                            oldLocusOfFocus, newLocusOfFocus)
+        if not newLocusOfFocus:
+            return
 
-        if self.utilities.isDocumentCell(newLocusOfFocus):
-            row, column, table = \
-                self.utilities.getRowColumnAndTable(newLocusOfFocus.parent)
+        cell = None
+        if self.utilities.isTextDocumentCell(newLocusOfFocus):
+            cell = newLocusOfFocus
+        elif self.utilities.isTextDocumentCell(newLocusOfFocus.parent):
+            cell = newLocusOfFocus.parent
+        if cell:
+            row, column = self.utilities.coordinatesForCell(cell)
             self.pointOfReference['lastRow'] = row
             self.pointOfReference['lastColumn'] = column
 
-    def onWindowActivated(self, event):
-        """Called whenever a property on an object changes.
-
-        Arguments:
-        - event: the Event
-        """
-
-        self.lastTextLength = -1
-        self.lastBadWord = ''
-        self.lastStartOff = -1
-        self.lastEndOff = -1
-
-        default.Script.onWindowActivated(self, event)
-
-        # Maybe it's the spellcheck dialog. Might as well try and see.
-        # If it is, we want to speak the misspelled word and context
-        # after we've spoken the window name.
-        if event.source \
-           and event.source.getRole() == pyatspi.ROLE_DIALOG \
-           and event.source.childCount \
-           and event.source[0].getRole() == pyatspi.ROLE_OPTION_PANE:
-            self.readMisspeltWord(event, event.source)
-
     def onNameChanged(self, event):
         """Called whenever a property on an object changes.
 
@@ -717,19 +587,8 @@ class Script(default.Script):
         - event: the Event
         """
 
-        # Check to see if if we've had a property-change event for the
-        # accessible name for the option pane in the spell check dialog.
-        # This (hopefully) means that the user has just corrected a
-        # spelling mistake, in which case, speak/braille the current
-        # misspelt word plus its context.
-        #
-        rolesList = [pyatspi.ROLE_OPTION_PANE, \
-                     pyatspi.ROLE_DIALOG, \
-                     pyatspi.ROLE_APPLICATION]
-        if self.utilities.hasMatchingHierarchy(event.source, rolesList) \
-           and self.utilities.isSameObject(
-                event.source.parent, orca_state.activeWindow):
-            self.readMisspeltWord(event, event.source)
+        if self.spellcheck.isCheckWindow(event.source):
+            return
 
         # Impress slide navigation.
         #
@@ -750,6 +609,11 @@ class Script(default.Script):
     def onActiveChanged(self, event):
         """Callback for object:state-changed:active accessibility events."""
 
+        if not event.source.parent:
+            msg = "SOFFICE: Event source lacks parent"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return
+
         # Prevent this events from activating the find operation.
         # See comment #18 of bug #354463.
         if self.findCommandRun:
@@ -768,52 +632,52 @@ class Script(default.Script):
         if self.utilities.isSameObject(event.any_data, orca_state.locusOfFocus):
             return
 
+        if event.source == self.spellcheck.getSuggestionsList():
+            if event.source.getState().contains(pyatspi.STATE_FOCUSED):
+                orca.setLocusOfFocus(event, event.any_data, False)
+                self.updateBraille(orca_state.locusOfFocus)
+                self.spellcheck.presentSuggestionListItem()
+            else:
+                self.spellcheck.presentErrorDetails()
+            return
+
+        if self.utilities.isSpreadSheetCell(event.any_data) \
+           and not event.any_data.getState().contains(pyatspi.STATE_FOCUSED) \
+           and not event.source.getState().contains(pyatspi.STATE_FOCUSED) :
+            msg = "SOFFICE: Neither source nor child have focused state. Clearing cache on table."
+            debug.println(debug.LEVEL_INFO, msg, True)
+            event.source.clearCache()
+
         default.Script.onActiveDescendantChanged(self, event)
 
-    def onChildrenChanged(self, event):
-        """Called when a child node has changed.
+    def onChildrenAdded(self, event):
+        """Callback for object:children-changed:add accessibility events."""
 
-        Arguments:
-        - event: the Event
-        """
-
-        try:
-            anyDataRole = event.any_data.getRole()
-        except:
+        if self.utilities.isSpreadSheetCell(event.any_data):
+            orca.setLocusOfFocus(event, event.any_data)
             return
 
-        if anyDataRole == pyatspi.ROLE_TABLE:
-            if self.utilities.isSpreadSheetCell(event.any_data, True):
-                orca.setLocusOfFocus(event, event.any_data)
-            return
-
-        if anyDataRole == pyatspi.ROLE_TABLE_CELL:
+        if self.utilities.isLastCell(event.any_data):
             activeRow = self.pointOfReference.get('lastRow', -1)
             activeCol = self.pointOfReference.get('lastColumn', -1)
             if activeRow < 0 or activeCol < 0:
                 return
 
-            eventRow, eventCol, table = \
-                self.utilities.getRowColumnAndTable(event.any_data)
-            try:
-                itable = table.queryTable()
-            except NotImplementedError:
-                return
+            if self.utilities.isDead(orca_state.locusOfFocus):
+                orca.setLocusOfFocus(event, event.source, False)
 
-            if eventRow == itable.nRows - 1 and eventCol == itable.nColumns - 1:
-                fullMessage = briefMessage = ""
-                voice = self.voices.get(settings.SYSTEM_VOICE)
-                if activeRow == itable.nRows:
-                    fullMessage = messages.TABLE_ROW_DELETED_FROM_END
-                    briefMessage = messages.TABLE_ROW_DELETED
-                else:
-                    fullMessage =  messages.TABLE_ROW_INSERTED_AT_END
-                    briefMessage = messages.TABLE_ROW_INSERTED
-                if fullMessage:
-                    self.presentMessage(fullMessage, briefMessage, voice)
-                return
+            self.utilities.handleUndoTextEvent(event)
+            rowCount, colCount = self.utilities.rowAndColumnCount(event.source)
+            if activeRow == rowCount:
+                full = messages.TABLE_ROW_DELETED_FROM_END
+                brief = messages.TABLE_ROW_DELETED
+            else:
+                full = messages.TABLE_ROW_INSERTED_AT_END
+                brief = messages.TABLE_ROW_INSERTED
+            self.presentMessage(full, brief)
+            return
 
-        default.Script.onChildrenChanged(self, event)
+        default.Script.onChildrenAdded(self, event)
 
     def onFocus(self, event):
         """Callback for focus: accessibility events."""
@@ -822,8 +686,22 @@ class Script(default.Script):
         # This callback remains just to handle bugs in applications and toolkits
         # during the remainder of the unstable (3.11) development cycle.
 
+        if self.utilities.isSameObject(orca_state.locusOfFocus, event.source):
+            return
+
+        if self.utilities.isFocusableLabel(event.source):
+            orca.setLocusOfFocus(event, event.source)
+            return
+
         role = event.source.getRole()
 
+        if self.utilities.isZombie(event.source) \
+           or role in [pyatspi.ROLE_TEXT, pyatspi.ROLE_LIST]:
+            comboBox = self.utilities.containingComboBox(event.source)
+            if comboBox:
+                orca.setLocusOfFocus(event, comboBox, True)
+                return
+
         # This seems to be something we inherit from Gtk+
         if role in [pyatspi.ROLE_TEXT, pyatspi.ROLE_PASSWORD_TEXT]:
             orca.setLocusOfFocus(event, event.source)
@@ -834,32 +712,55 @@ class Script(default.Script):
             orca.setLocusOfFocus(event, event.source)
             return
 
+        # Ditto.
+        if role == pyatspi.ROLE_TOGGLE_BUTTON:
+            orca.setLocusOfFocus(event, event.source)
+            return
+
         # Ditto.
         if role == pyatspi.ROLE_COMBO_BOX:
             orca.setLocusOfFocus(event, event.source)
             return
 
+        # Ditto.
+        if role == pyatspi.ROLE_PANEL and event.source.name:
+            orca.setLocusOfFocus(event, event.source)
+            return
+
     def onFocusedChanged(self, event):
         """Callback for object:state-changed:focused accessibility events."""
 
-        if self.isStructuralNavigationCommand():
+        if self._inSayAll:
+            return
+
+        if self._lastCommandWasStructNav:
             return
 
         if not event.detail1:
             return
 
-        if event.source.getRoleName() == 'text frame':
+        if self.utilities.isAnInputLine(event.source):
+            msg = "SOFFICE: Event ignored: spam from inputLine"
+            debug.println(debug.LEVEL_INFO, msg, True)
             return
+        if event.source.childCount and self.utilities.isAnInputLine(event.source[0]):
+            msg = "SOFFICE: Event ignored: spam from inputLine parent"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return
+
+        role = event.source.getRole()
+        if role in [pyatspi.ROLE_TEXT, pyatspi.ROLE_LIST]:
+            comboBox = self.utilities.containingComboBox(event.source)
+            if comboBox:
+                orca.setLocusOfFocus(event, comboBox, True)
+                return
 
         parent = event.source.parent
-        if parent and parent.getRoleName() == 'text frame':
-            return
-
         if parent and parent.getRole() == pyatspi.ROLE_TOOL_BAR:
             default.Script.onFocusedChanged(self, event)
             return
 
-        role = event.source.getRole()
+        # TODO - JD: Verify this is still needed
         ignoreRoles = [pyatspi.ROLE_FILLER, pyatspi.ROLE_PANEL]
         if role in ignoreRoles:
             return
@@ -868,70 +769,64 @@ class Script(default.Script):
         if role == pyatspi.ROLE_MENU:
             return
 
-        obj, offset = self.pointOfReference.get("lastCursorPosition", (None, -1))
-        textSelections = self.pointOfReference.get('textSelections', {})
-        start, end = textSelections.get(hash(obj), (0, 0))
-        if start != end:
-            return
-
         if self.utilities._flowsFromOrToSelection(event.source):
             return
 
         if role == pyatspi.ROLE_PARAGRAPH:
+            obj, offset = self.pointOfReference.get("lastCursorPosition", (None, -1))
+            start, end, string = self.utilities.getCachedTextSelection(obj)
+            if start != end:
+                return
+
             keyString, mods = self.utilities.lastKeyAndModifiers()
             if keyString in ["Left", "Right"]:
                 orca.setLocusOfFocus(event, event.source, False)
                 return
 
-        # We should present this in response to active-descendant-changed events
-        if event.source.getState().contains(pyatspi.STATE_MANAGES_DESCENDANTS):
-            return
+        if self.utilities.isSpreadSheetTable(event.source) and orca_state.locusOfFocus:
+            if self.utilities.isDead(orca_state.locusOfFocus):
+                msg = "SOFFICE: Event believed to be post-editing focus claim. Dead locusOfFocus."
+                debug.println(debug.LEVEL_INFO, msg, True)
+                orca.setLocusOfFocus(event, event.source, False)
+                return
+
+            if  orca_state.locusOfFocus.getRole() in [pyatspi.ROLE_PARAGRAPH, pyatspi.ROLE_TABLE_CELL]:
+                msg = "SOFFICE: Event believed to be post-editing focus claim based on role."
+                debug.println(debug.LEVEL_INFO, msg, True)
+                orca.setLocusOfFocus(event, event.source, False)
+                return
 
         default.Script.onFocusedChanged(self, event)
 
     def onCaretMoved(self, event):
-        """Called whenever the caret moves.
-
-        Arguments:
-        - event: the Event
-        """
-
-        if self.isStructuralNavigationCommand():
-            return
+        """Callback for object:text-caret-moved accessibility events."""
 
         if event.detail1 == -1:
             return
 
-        if self.utilities.isCellBeingEdited(event.source):
-            orca.setLocusOfFocus(event, event.source.parent, False)
-
-        if not orca_state.locusOfFocus:
-            default.Script.onCaretMoved(self, event)
-            return
-
-        if orca_state.locusOfFocus.getRole() == pyatspi.ROLE_TABLE_CELL:
-            default.Script.onCaretMoved(self, event)
-            return
-
-        # The lists and combo boxes in the Formatting toolbar emit
-        # object:active-descendant-changed events which cause us
-        # to set the locusOfFocus to the list item. If the user then
-        # arrows within the text portion, we will not present it due
-        # to the event not being from the locusOfFocus. A similar
-        # issue is present in the Target entry of the Hyperlink dialog
-        # for OOo 3.2.
-        #
-        if event.source.getRole() == pyatspi.ROLE_TEXT \
-           and self.utilities.ancestorWithRole(
-               event.source,
-               [pyatspi.ROLE_TOOL_BAR, pyatspi.ROLE_DIALOG],
-               [pyatspi.ROLE_FRAME]):
-            orca.setLocusOfFocus(event, event.source, False)
+        if event.source.getRole() == pyatspi.ROLE_PARAGRAPH \
+           and not event.source.getState().contains(pyatspi.STATE_FOCUSED):
+            event.source.clearCache()
+            if event.source.getState().contains(pyatspi.STATE_FOCUSED):
+                msg = "SOFFICE: Clearing cache was needed due to missing state-changed event."
+                debug.println(debug.LEVEL_INFO, msg, True)
 
         if self.utilities._flowsFromOrToSelection(event.source):
+           return
+
+        if self._lastCommandWasStructNav:
             return
 
-        default.Script.onCaretMoved(self, event)
+        if self.utilities.isSpreadSheetCell(orca_state.locusOfFocus):
+            msg = "SOFFICE: locusOfFocus %s is spreadsheet cell" % orca_state.locusOfFocus
+            debug.println(debug.LEVEL_INFO, msg, True)
+
+            if not self.utilities.isCellBeingEdited(event.source):
+                msg = "SOFFICE: Event ignored: Source is not cell being edited."
+                debug.println(debug.LEVEL_INFO, msg, True)
+                return
+
+        super().onCaretMoved(event)
 
     def onCheckedChanged(self, event):
         """Callback for object:state-changed:checked accessibility events."""
@@ -944,6 +839,11 @@ class Script(default.Script):
             default.Script.onCheckedChanged(self, event)
             return
  
+        sourceWindow = self.utilities.topLevelObject(obj)
+        focusWindow = self.utilities.topLevelObject(orca_state.locusOfFocus)
+        if sourceWindow != focusWindow:
+            return
+
         # Announce when the toolbar buttons are toggled if we just toggled
         # them; not if we navigated to some text.
         weToggledIt = False
@@ -962,40 +862,72 @@ class Script(default.Script):
         if weToggledIt:
             speech.speak(self.speechGenerator.generateSpeech(obj))
 
-    def onRowReordered(self, event):
-        """Callback for object:row-reordered accessibility events."""
+    def onSelectedChanged(self, event):
+        """Callback for object:state-changed:selected accessibility events."""
 
-        # We're seeing a crazy ton of these emitted bogusly.
-        pass
+        full, brief = "", ""
+        if self.utilities.isSelectedTextDeletionEvent(event):
+            msg = "SOFFICE: Change is believed to be due to deleting selected text"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            full = messages.SELECTION_DELETED
+        elif self.utilities.isSelectedTextRestoredEvent(event):
+            msg = "SOFFICE: Selection is believed to be due to restoring selected text"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            if self.utilities.handleUndoTextEvent(event):
+                full = messages.SELECTION_RESTORED
 
-    def onTextAttributesChanged(self, event):
-        """Callback for object:text-attributes-changed accessibility events."""
+        if full or brief:
+            self.presentMessage(full, brief)
+            self.utilities.updateCachedTextSelection(event.source)
+            return
 
-        # LibreOffice emits this signal nearly every time text is typed,
-        # even though the text attributes haven't changed:
-        # https://bugs.freedesktop.org/show_bug.cgi?id=71556
+        super().onSelectedChanged(event)
 
-        # LibreOffice fails to emit this signal the main time we are looking
-        # for it, namely to present that a misspelled word was typed:
-        # https://bugs.freedesktop.org/show_bug.cgi?id=71558
+    def onSelectionChanged(self, event):
+        """Callback for object:selection-changed accessibility events."""
 
-        # Useless signals are useless.
-        pass
+        if self.utilities.isSpreadSheetTable(event.source):
+            if _settingsManager.getSetting('onlySpeakDisplayedText'):
+                return
+            if _settingsManager.getSetting('alwaysSpeakSelectedSpreadsheetRange'):
+                self.utilities.speakSelectedCellRange(event.source)
+                return
+            if self.utilities.handleRowAndColumnSelectionChange(event.source):
+                return
+            self.utilities.handleCellSelectionChange(event.source)
+            return
 
-    def getTextLineAtCaret(self, obj, offset=None):
-        """Gets the line of text where the caret is. Overridden here to
-        handle combo boxes who have a text object with a caret offset
-        of -1 as a child.
+        if not self.utilities.isComboBoxSelectionChange(event):
+            super().onSelectionChanged(event)
+            return
 
-        Argument:
-        - obj: an Accessible object that implements the AccessibleText
-          interface
-        - offset: an optional caret offset to use. (Not used here at the
-          moment, but needed in the Gecko script)
+        selectedChildren = self.utilities.selectedChildren(event.source)
+        if len(selectedChildren) == 1 \
+           and self.utilities.containingComboBox(event.source) == \
+               self.utilities.containingComboBox(orca_state.locusOfFocus):
+            orca.setLocusOfFocus(event, selectedChildren[0], True)
 
-        Returns the [string, caretOffset, startOffset] for the line of text
-        where the caret is.
-        """
+    def onTextSelectionChanged(self, event):
+        """Callback for object:text-selection-changed accessibility events."""
+
+        if self.utilities.isComboBoxNoise(event):
+            msg = "SOFFICE: Event is believed to be combo box noise"
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return
+
+        if self.utilities.isDead(event.source):
+            msg = "SOFFICE: Ignoring event from dead source."
+            debug.println(debug.LEVEL_INFO, msg, True)
+            return
+
+        if event.source != orca_state.locusOfFocus \
+           and event.source.getState().contains(pyatspi.STATE_FOCUSED):
+            orca.setLocusOfFocus(event, event.source, False)
+
+        super().onTextSelectionChanged(event)
+
+    def getTextLineAtCaret(self, obj, offset=None, startOffset=None, endOffset=None):
+        """To-be-removed. Returns the string, caretOffset, startOffset."""
 
         if obj.parent.getRole() == pyatspi.ROLE_COMBO_BOX:
             try:
@@ -1014,8 +946,39 @@ class Script(default.Script):
 
                 return [lineString, 0, startOffset]
 
-        textLine = default.Script.getTextLineAtCaret(self, obj, offset)
+        textLine = super().getTextLineAtCaret(obj, offset, startOffset, endOffset)
         if not obj.getState().contains(pyatspi.STATE_FOCUSED):
             textLine[0] = self.utilities.displayedText(obj)
 
         return textLine
+
+    def onWindowActivated(self, event):
+        """Callback for window:activate accessibility events."""
+
+        super().onWindowActivated(event)
+        if not self.spellcheck.isCheckWindow(event.source):
+            return
+
+        if event.source[0].getRole() == pyatspi.ROLE_DIALOG:
+            orca.setLocusOfFocus(event, event.source[0], False)
+
+        self.spellcheck.presentErrorDetails()
+
+    def onWindowDeactivated(self, event):
+        """Callback for window:deactivate accessibility events."""
+
+        self._lastCommandWasStructNav = False
+
+        super().onWindowDeactivated(event)
+        self.spellcheck.deactivate()
+
+    def whereAmISelection(self, inputEvent=None, obj=None):
+        obj = obj or orca_state.locusOfFocus
+        if not self.utilities.isSpreadSheetCell(obj):
+            if self.utilities.inDocumentContent(obj):
+                # Because for some reason, the document implements the selection
+                # interface as if it were a spreadsheet or listbox. *sigh*
+                return super()._whereAmISelectedText(inputEvent, obj)
+            return super().whereAmISelection(inputEvent, obj)
+
+        return self.utilities.speakSelectedCellRange(self.utilities.getTable(obj))
