@@ -33,6 +33,7 @@ import pyatspi.utils as utils
 import orca.scripts.default as default
 import orca.cmdnames as cmdnames
 import orca.debug as debug
+import orca.eventsynthesizer as eventsynthesizer
 import orca.guilabels as guilabels
 import orca.input_event as input_event
 import orca.messages as messages
@@ -42,8 +43,8 @@ import orca.settings_manager as settings_manager
 import orca.speechserver as speechserver
 import orca.orca_state as orca_state
 import orca.speech as speech
+import orca.structural_navigation as structural_navigation
 
-from .structural_navigation import StructuralNavigation
 from .braille_generator import BrailleGenerator
 from .speech_generator import SpeechGenerator
 from .script_utilities import Utilities
@@ -58,8 +59,6 @@ _settingsManager = settings_manager.getManager()
 
 class Script(default.Script):
 
-    CARET_NAVIGATION_KEYS = ['Left', 'Right', 'Up', 'Down', 'Home', 'End']
-
     def __init__(self, app):
         """Creates a new script for WebKitGtk applications.
 
@@ -72,7 +71,7 @@ class Script(default.Script):
         self._lastCaretContext = None, -1
         self.sayAllOnLoadCheckButton = None
 
-        if _settingsManager.getSetting('sayAllOnLoad') == None:
+        if _settingsManager.getSetting('sayAllOnLoad') is None:
             _settingsManager.setSetting('sayAllOnLoad', True)
 
     def setupInputEventHandlers(self):
@@ -140,38 +139,33 @@ class Script(default.Script):
 
         return SpeechGenerator(self)
 
-    def getStructuralNavigation(self):
-        """Returns the 'structural navigation' class for this script."""
-
-        types = self.getEnabledStructuralNavigationTypes()
-        return StructuralNavigation(self, types, True)
-
     def getEnabledStructuralNavigationTypes(self):
         """Returns a list of the structural navigation object types
         enabled in this script."""
 
-        enabledTypes = [StructuralNavigation.BLOCKQUOTE,
-                        StructuralNavigation.BUTTON,
-                        StructuralNavigation.CHECK_BOX,
-                        StructuralNavigation.CHUNK,
-                        StructuralNavigation.COMBO_BOX,
-                        StructuralNavigation.ENTRY,
-                        StructuralNavigation.FORM_FIELD,
-                        StructuralNavigation.HEADING,
-                        StructuralNavigation.LANDMARK,
-                        StructuralNavigation.LINK,
-                        StructuralNavigation.LIST,
-                        StructuralNavigation.LIST_ITEM,
-                        StructuralNavigation.LIVE_REGION,
-                        StructuralNavigation.PARAGRAPH,
-                        StructuralNavigation.RADIO_BUTTON,
-                        StructuralNavigation.SEPARATOR,
-                        StructuralNavigation.TABLE,
-                        StructuralNavigation.TABLE_CELL,
-                        StructuralNavigation.UNVISITED_LINK,
-                        StructuralNavigation.VISITED_LINK]
-
-        return enabledTypes
+        return [structural_navigation.StructuralNavigation.BLOCKQUOTE,
+                structural_navigation.StructuralNavigation.BUTTON,
+                structural_navigation.StructuralNavigation.CHECK_BOX,
+                structural_navigation.StructuralNavigation.CHUNK,
+                structural_navigation.StructuralNavigation.CLICKABLE,
+                structural_navigation.StructuralNavigation.COMBO_BOX,
+                structural_navigation.StructuralNavigation.CONTAINER,
+                structural_navigation.StructuralNavigation.ENTRY,
+                structural_navigation.StructuralNavigation.FORM_FIELD,
+                structural_navigation.StructuralNavigation.HEADING,
+                structural_navigation.StructuralNavigation.IMAGE,
+                structural_navigation.StructuralNavigation.LANDMARK,
+                structural_navigation.StructuralNavigation.LINK,
+                structural_navigation.StructuralNavigation.LIST,
+                structural_navigation.StructuralNavigation.LIST_ITEM,
+                structural_navigation.StructuralNavigation.LIVE_REGION,
+                structural_navigation.StructuralNavigation.PARAGRAPH,
+                structural_navigation.StructuralNavigation.RADIO_BUTTON,
+                structural_navigation.StructuralNavigation.SEPARATOR,
+                structural_navigation.StructuralNavigation.TABLE,
+                structural_navigation.StructuralNavigation.TABLE_CELL,
+                structural_navigation.StructuralNavigation.UNVISITED_LINK,
+                structural_navigation.StructuralNavigation.VISITED_LINK]
 
     def getUtilities(self):
         """Returns the utilites for this script."""
@@ -179,16 +173,14 @@ class Script(default.Script):
         return Utilities(self)
 
     def onCaretMoved(self, event):
-        """Called whenever the caret moves.
-
-        Arguments:
-        - event: the Event
-        """
+        """Callback for object:text-caret-moved accessibility events."""
 
         if self._inSayAll:
             return
 
-        self._lastCaretContext = event.source, event.detail1
+        if not self.utilities.isWebKitGtk(event.source):
+            super().onCaretMoved(event)
+            return
 
         lastKey, mods = self.utilities.lastKeyAndModifiers()
         if lastKey in ['Tab', 'ISO_Left_Tab']:
@@ -201,9 +193,8 @@ class Script(default.Script):
             self.updateBraille(event.source)
             return
 
-        orca.setLocusOfFocus(event, event.source, False)
-
-        default.Script.onCaretMoved(self, event)
+        self.utilities.setCaretContext(event.source, event.detail1)
+        super().onCaretMoved(event)
 
     def onDocumentReload(self, event):
         """Callback for document:reload accessibility events."""
@@ -223,7 +214,7 @@ class Script(default.Script):
         # is grabbed rather than set the caret at the start. But for simple
         # content in both Yelp and Epiphany this is alright for now.
         obj, offset = self.utilities.setCaretAtStart(event.source)
-        orca.setLocusOfFocus(event, obj, False)
+        self.utilities.setCaretContext(obj, offset)
 
         self.updateBraille(obj)
         if _settingsManager.getSetting('sayAllOnLoad') \
@@ -239,38 +230,29 @@ class Script(default.Script):
     def onFocusedChanged(self, event):
         """Callback for object:state-changed:focused accessibility events."""
 
-        if self._inSayAll:
+        if self._inSayAll or not event.detail1:
+            return
+
+        if not self.utilities.isWebKitGtk(event.source):
+            super().onFocusedChanged(event)
+            return
+
+        contextObj, offset = self.utilities.getCaretContext()
+        if event.source == contextObj:
             return
 
         obj = event.source
         role = obj.getRole()
-        self._lastCaretContext = obj, -1
-
-        if role == pyatspi.ROLE_LIST_ITEM and obj.childCount:
-            return
-
-        widgetRoles = [pyatspi.ROLE_MENU,
-                       pyatspi.ROLE_MENU_ITEM,
-                       pyatspi.ROLE_LIST_ITEM,
-                       pyatspi.ROLE_RADIO_BUTTON]
-        if role in widgetRoles:
-            default.Script.onFocusedChanged(self, event)
-            return
-
         textRoles = [pyatspi.ROLE_HEADING,
                      pyatspi.ROLE_PANEL,
                      pyatspi.ROLE_PARAGRAPH,
                      pyatspi.ROLE_SECTION,
                      pyatspi.ROLE_TABLE_CELL]
-        if role in textRoles:
+        if role in textRoles \
+           or (role == pyatspi.ROLE_LIST_ITEM and obj.childCount):
             return
 
-        if not (role == pyatspi.ROLE_LINK and obj.childCount):
-            lastKey, mods = self.utilities.lastKeyAndModifiers()
-            if lastKey in self.CARET_NAVIGATION_KEYS:
-                return
-
-        default.Script.onFocusedChanged(self, event)
+        super().onFocusedChanged(event)
 
     def onBusyChanged(self, event):
         """Callback for object:state-changed:busy accessibility events."""
@@ -311,6 +293,8 @@ class Script(default.Script):
             else:
                 speech.speak(self.speechGenerator.generateSpeech(obj))
 
+        self.pointOfReference["lastTextUnitSpoken"] = "char"
+
     def sayWord(self, obj):
         """Speaks the word at the caret.
 
@@ -327,6 +311,8 @@ class Script(default.Script):
         for (obj, start, end, string) in objects:
             self.sayPhrase(obj, start, end)
 
+        self.pointOfReference["lastTextUnitSpoken"] = "word"
+
     def sayLine(self, obj):
         """Speaks the line at the caret.
 
@@ -352,6 +338,8 @@ class Script(default.Script):
             if obj.getRole() in rolesToSpeak:
                 speech.speak(self.speechGenerator.getRoleName(obj))
 
+        self.pointOfReference["lastTextUnitSpoken"] = "line"
+
     def sayPhrase(self, obj, startOffset, endOffset):
         """Speaks the text of an Accessible object between the given offsets.
 
@@ -367,11 +355,7 @@ class Script(default.Script):
 
         phrase = self.utilities.substring(obj, startOffset, endOffset)
         if len(phrase) and phrase != "\n":
-            if phrase.isupper():
-                voice = self.voices[settings.UPPERCASE_VOICE]
-            else:
-                voice = self.voices[settings.DEFAULT_VOICE]
-
+            voice = self.speechGenerator.voice(string=phrase)
             phrase = self.utilities.adjustForRepeats(phrase)
             links = [x for x in obj if x.getRole() == pyatspi.ROLE_LINK]
             if links:
@@ -382,6 +366,8 @@ class Script(default.Script):
             #
             self.sayCharacter(obj)
 
+        self.pointOfReference["lastTextUnitSpoken"] = "phrase"
+
     def skipObjectEvent(self, event):
         """Gives us, and scripts, the ability to decide an event isn't
         worth taking the time to process under the current circumstances.
@@ -534,8 +520,7 @@ class Script(default.Script):
         if obj.getRole() == pyatspi.ROLE_LINK:
             obj = obj.parent
 
-        docRoles = [pyatspi.ROLE_DOCUMENT_FRAME, pyatspi.ROLE_DOCUMENT_WEB]
-        document = utils.findAncestor(obj, lambda x: x.getRole() in docRoles)
+        document = self.utilities.getDocumentForObject(obj)
         if not document or document.getState().contains(pyatspi.STATE_BUSY):
             return
 
@@ -551,18 +536,21 @@ class Script(default.Script):
         if sayAllStyle == settings.SAYALL_STYLE_SENTENCE:
             boundary = pyatspi.TEXT_BOUNDARY_SENTENCE_START
 
+        voices = _settingsManager.getSetting('voices')
+        systemVoice = voices.get(settings.SYSTEM_VOICE)
+
         self._inSayAll = True
         offset = textObjs[0].queryText().caretOffset
         for textObj in textObjs:
             textSegments = self.getTextSegments(textObj, boundary, offset)
-            roleInfo = self.speechGenerator.getRoleName(textObj)
-            if roleInfo:
-                roleName, voice = roleInfo
-                textSegments.append([roleName, 0, -1, voice])
+            roleName = self.speechGenerator.getRoleName(textObj)
+            if roleName:
+                textSegments.append([roleName, 0, -1, systemVoice])
 
             for (string, start, end, voice) in textSegments:
                 context = speechserver.SayAllContext(textObj, string, start, end)
                 self._sayAllContexts.append(context)
+                eventsynthesizer.scrollIntoView(obj, start, end)
                 yield [context, voice]
 
             offset = 0
@@ -572,6 +560,8 @@ class Script(default.Script):
 
     def __sayAllProgressCallback(self, context, progressType):
         if progressType == speechserver.SayAllContext.PROGRESS:
+            orca.emitRegionChanged(
+                context.obj, context.currentOffset, context.currentEndOffset, orca.SAY_ALL)
             return
 
         obj = context.obj
@@ -593,6 +583,7 @@ class Script(default.Script):
             self._sayAllContexts = []
             if not self._lastCommandWasStructNav:
                 text.setCaretOffset(offset)
+            orca.emitRegionChanged(obj, offset)
             return
 
         # SayAllContext.COMPLETED doesn't necessarily mean done with SayAll;
@@ -609,6 +600,7 @@ class Script(default.Script):
             if [l for l in links if l.startIndex <= offset <= l.endIndex]:
                 return
 
+        orca.emitRegionChanged(obj, offset, mode=orca.SAY_ALL)
         text.setCaretOffset(offset)
 
     def getTextLineAtCaret(self, obj, offset=None, startOffset=None, endOffset=None):
@@ -630,17 +622,16 @@ class Script(default.Script):
 
         return textLine
 
-    def updateBraille(self, obj, extraRegion=None):
+    def updateBraille(self, obj, **args):
         """Updates the braille display to show the given object.
 
         Arguments:
         - obj: the Accessible
-        - extra: extra Region to add to the end
         """
 
         if not _settingsManager.getSetting('enableBraille') \
            and not _settingsManager.getSetting('enableBrailleMonitor'):
-            debug.println(debug.LEVEL_INFO, "BRAILLE: update disabled")
+            debug.println(debug.LEVEL_INFO, "BRAILLE: update disabled", True)
             return
 
         if not obj:
@@ -649,7 +640,7 @@ class Script(default.Script):
         if not self.utilities.isWebKitGtk(obj) \
            or (not self.utilities.isInlineContainer(obj) \
                and not self.utilities.isTextListItem(obj)):
-            default.Script.updateBraille(self, obj, extraRegion)
+            default.Script.updateBraille(self, obj, **args)
             return
 
         brailleLine = self.getNewBrailleLine(clearBraille=True, addLine=True)
@@ -665,6 +656,7 @@ class Script(default.Script):
             self.addBrailleRegionsToLine(regions, brailleLine)
             self.setBrailleFocus(fRegion)
 
+        extraRegion = args.get('extraRegion')
         if extraRegion:
             self.addBrailleRegionToLine(extraRegion, brailleLine)
 
