@@ -34,6 +34,7 @@ from orca import debug
 from orca import messages
 from orca import object_properties
 from orca import orca_state
+from orca import settings
 from orca import settings_manager
 from orca import speech_generator
 
@@ -45,6 +46,26 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
     def __init__(self, script):
         super().__init__(script)
 
+    def _generateOldAncestors(self, obj, **args):
+        if args.get('index', 0) > 0:
+            return []
+
+        priorObj = args.get('priorObj')
+        if self._script.utilities.isInlineIframeDescendant(priorObj):
+            return []
+
+        return super()._generateOldAncestors(obj, **args)
+
+    def _generateNewAncestors(self, obj, **args):
+        if args.get('index', 0) > 0 \
+           and not self._script.utilities.isListDescendant(obj):
+            return []
+
+        if self._script.utilities.isInlineIframeDescendant(obj):
+            return []
+
+        return super()._generateNewAncestors(obj, **args)
+
     def _generateAncestors(self, obj, **args):
         if not self._script.utilities.inDocumentContent(obj):
             return super()._generateAncestors(obj, **args)
@@ -62,7 +83,8 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
 
         if self._script.utilities.isLink(obj) \
            or self._script.utilities.isLandmark(obj) \
-           or self._script.utilities.isMath(obj):
+           or self._script.utilities.isMath(obj) \
+           or obj.getRole() in [pyatspi.ROLE_TOOL_TIP, pyatspi.ROLE_STATUS_BAR]:
             return result
 
         args['stopAtRoles'] = [pyatspi.ROLE_DOCUMENT_FRAME,
@@ -70,14 +92,14 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                                pyatspi.ROLE_EMBEDDED,
                                pyatspi.ROLE_INTERNAL_FRAME,
                                pyatspi.ROLE_MATH,
-                               pyatspi.ROLE_MENU_BAR,
-                               pyatspi.ROLE_TOOL_BAR]
+                               pyatspi.ROLE_MENU_BAR]
         args['skipRoles'] = [pyatspi.ROLE_PARAGRAPH,
                              pyatspi.ROLE_HEADING,
                              pyatspi.ROLE_LABEL,
                              pyatspi.ROLE_LINK,
                              pyatspi.ROLE_LIST_ITEM,
                              pyatspi.ROLE_TEXT]
+        args['stopAfterRoles'] = [pyatspi.ROLE_TOOL_BAR]
 
         if self._script.utilities.isEditableDescendantOfComboBox(obj):
             args['skipRoles'].append(pyatspi.ROLE_COMBO_BOX)
@@ -104,7 +126,35 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         # call utility methods rather than generate it.
         return super()._generateAnyTextSelection(obj, **args)
 
+    def _generateHasPopup(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
+        if not self._script.utilities.inDocumentContent(obj):
+            return []
+
+        result = []
+        popupType = self._script.utilities.popupType(obj)
+        if popupType == 'dialog':
+            result = [messages.HAS_POPUP_DIALOG]
+        elif popupType == 'grid':
+            result = [messages.HAS_POPUP_GRID]
+        elif popupType == 'listbox':
+            result = [messages.HAS_POPUP_LISTBOX]
+        elif popupType in ('menu', 'true'):
+            result = [messages.HAS_POPUP_MENU]
+        elif popupType == 'tree':
+            result = [messages.HAS_POPUP_TREE]
+
+        if result:
+            result.extend(self.voice(speech_generator.SYSTEM))
+
+        return result
+
     def _generateClickable(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if not self._script.utilities.inDocumentContent(obj):
             return []
 
@@ -120,6 +170,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return []
 
     def _generateDescription(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if not self._script.utilities.inDocumentContent(obj):
             return super()._generateDescription(obj, **args)
 
@@ -150,6 +203,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return super()._generateDescription(obj, **args)
 
     def _generateHasLongDesc(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if not self._script.utilities.inDocumentContent(obj):
             return []
 
@@ -164,23 +220,136 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
 
         return []
 
+    def _generateHasDetails(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
+        if not self._script.utilities.inDocumentContent(obj):
+            return super()._generateHasDetails(obj, **args)
+
+        objs = self._script.utilities.detailsIn(obj)
+        if not objs:
+            return []
+
+        objString = lambda x: str.strip("%s %s" % (x.name, self.getLocalizedRoleName(x)))
+        toPresent = ", ".join(set(map(objString, objs)))
+
+        args['stringType'] = 'hasdetails'
+        result = [self._script.formatting.getString(**args) % toPresent]
+        result.extend(self.voice(speech_generator.SYSTEM))
+        return result
+
+    def _generateAllDetails(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
+        objs = self._script.utilities.detailsIn(obj)
+        if not objs:
+            container = pyatspi.findAncestor(obj, self._script.utilities.hasDetails)
+            objs = self._script.utilities.detailsIn(container)
+
+        if not objs:
+            return []
+
+        args['stringType'] = 'hasdetails'
+        result = [self._script.formatting.getString(**args) % ""]
+        result.extend(self.voice(speech_generator.SYSTEM))
+
+        result = []
+        for o in objs:
+            result.append(self.getLocalizedRoleName(o))
+            result.extend(self.voice(speech_generator.SYSTEM))
+
+            string = self._script.utilities.expandEOCs(o)
+            if not string.strip():
+                continue
+
+            result.append(string)
+            result.extend(self.voice(speech_generator.DEFAULT))
+            result.extend(self._generatePause(o))
+
+        return result
+
+    def _generateDetailsFor(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
+        if not self._script.utilities.inDocumentContent(obj):
+            return super()._generateDetailsFor(obj, **args)
+
+        objs = self._script.utilities.detailsFor(obj)
+        if not objs:
+            return []
+
+        if args.get('leaving'):
+            return []
+
+        lastKey, mods = self._script.utilities.lastKeyAndModifiers()
+        if (lastKey in ['Down', 'Right'] or self._script.inSayAll()) and args.get('startOffset'):
+            return []
+        if lastKey in ['Up', 'Left']:
+            text = self._script.utilities.queryNonEmptyText(obj)
+            if text and args.get('endOffset') not in [None, text.characterCount]:
+                return []
+
+        result = []
+        objArgs = {'stringType': 'detailsfor', 'mode': args.get('mode')}
+        for o in objs:
+            string = self._script.utilities.displayedText(o) or self.getLocalizedRoleName(o)
+            words = string.split()
+            if len(words) > 5:
+                words = words[0:5] + ['...']
+
+            result.append(self._script.formatting.getString(**objArgs) % " ".join(words))
+            result.extend(self.voice(speech_generator.SYSTEM))
+            result.extend(self._generatePause(o, **objArgs))
+
+        return result
+
     def _generateLabelOrName(self, obj, **args):
         if not self._script.utilities.inDocumentContent(obj):
             return super()._generateLabelOrName(obj, **args)
 
         if self._script.utilities.isTextBlockElement(obj) \
            and not self._script.utilities.isLandmark(obj) \
-           and not self._script.utilities.isDPub(obj):
+           and not self._script.utilities.isDocument(obj) \
+           and not self._script.utilities.isDPub(obj) \
+           and not self._script.utilities.isContentSuggestion(obj):
+            return []
+
+        priorObj = args.get("priorObj")
+        if obj == priorObj:
+            return []
+
+        if priorObj and priorObj in self._script.utilities.labelsForObject(obj):
+            return []
+
+        if self._script.utilities.isContentEditableWithEmbeddedObjects(obj) \
+           or self._script.utilities.isDocument(obj):
+            lastKey, mods = self._script.utilities.lastKeyAndModifiers()
+            if lastKey in ["Home", "End", "Up", "Down", "Left", "Right", "Page_Up", "Page_Down"]:
+                return []
+
+        if priorObj and priorObj.getRole() == pyatspi.ROLE_PAGE_TAB and priorObj.name == obj.name:
             return []
 
         if obj.name:
             name = obj.name
             if not self._script.utilities.hasExplicitName(obj):
                 name = name.strip()
+
+            if self._script.utilities.shouldVerbalizeAllPunctuation(obj):
+                name = self._script.utilities.verbalizeAllPunctuation(name)
+
             result = [name]
             result.extend(self.voice(speech_generator.DEFAULT))
             return result
 
+        if obj.getRole() == pyatspi.ROLE_CHECK_BOX:
+            gridCell = pyatspi.findAncestor(obj, self._script.utilities.isGridCell)
+            if gridCell:
+                return super()._generateLabelOrName(gridCell, **args)
+
         return super()._generateLabelOrName(obj, **args)
 
     def _generateName(self, obj, **args):
@@ -189,32 +358,34 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
 
         if self._script.utilities.isTextBlockElement(obj) \
            and not self._script.utilities.isLandmark(obj) \
-           and not self._script.utilities.isDPub(obj):
+           and not self._script.utilities.isDPub(obj) \
+           and not args.get('inFlatReview'):
             return []
 
-        if obj.name and not self._script.utilities.hasValidName(obj):
+        if self._script.utilities.hasVisibleCaption(obj):
             return []
 
+        if self._script.utilities.isFigure(obj) and args.get('ancestorOf'):
+            caption = args.get('ancestorOf')
+            if caption.getRole() != pyatspi.ROLE_CAPTION:
+                isCaption = lambda x: x and x.getRole() == pyatspi.ROLE_CAPTION
+                caption = pyatspi.findAncestor(caption, isCaption)
+            if caption and hash(obj) in self._script.utilities.labelTargets(caption):
+                return []
+
         role = args.get('role', obj.getRole())
-        alwaysPresent = [pyatspi.ROLE_PUSH_BUTTON,
-                         pyatspi.ROLE_IMAGE]
-
-        if obj.parent and obj.name and obj.name == obj.parent.name \
-           and obj != orca_state.locusOfFocus and role not in alwaysPresent:
-            return []
 
         # TODO - JD: Once the formatting strings are vastly cleaned up
         # or simply removed, hacks like this won't be needed.
         if role in [pyatspi.ROLE_COMBO_BOX, pyatspi.ROLE_SPIN_BUTTON]:
             return super()._generateName(obj, **args)
 
-        if self._script.utilities.isLink(obj) \
-           and not self._script.utilities.hasExplicitName(obj):
-            return []
-
         if obj.name:
             if self._script.utilities.preferDescriptionOverName(obj):
                 result = [obj.description]
+            elif self._script.utilities.isLink(obj) \
+                 and not self._script.utilities.hasExplicitName(obj):
+                return []
             else:
                 name = obj.name
                 if not self._script.utilities.hasExplicitName(obj):
@@ -242,6 +413,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return super()._generateLabel(obj, **args)
 
     def _generateNewNodeLevel(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if self._script.utilities.isTextBlockElement(obj) \
            or self._script.utilities.isLink(obj):
             return []
@@ -249,6 +423,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return super()._generateNewNodeLevel(obj, **args)
 
     def _generateLeaving(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if not args.get('leaving'):
             return []
 
@@ -267,7 +444,8 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return []
 
     def _generateNumberOfChildren(self, obj, **args):
-        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+        if _settingsManager.getSetting('onlySpeakDisplayedText') \
+           or _settingsManager.getSetting('speechVerbosityLevel') == settings.VERBOSITY_LEVEL_BRIEF:
             return []
 
         # We handle things even for non-document content due to issues in
@@ -310,9 +488,15 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return self._generateDisplayedText(obj, **args)
 
     def _generateRoleName(self, obj, **args):
+        if _settingsManager.getSetting('onlySpeakDisplayedText'):
+            return []
+
         if not self._script.utilities.inDocumentContent(obj):
             return super()._generateRoleName(obj, **args)
 
+        if obj == args.get('priorObj'):
+            return []
+
         result = []
         acss = self.voice(speech_generator.SYSTEM)
 
@@ -349,6 +533,8 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
             doNotSpeak.append(pyatspi.ROLE_TABLE_CELL)
             doNotSpeak.append(pyatspi.ROLE_TEXT)
             doNotSpeak.append(pyatspi.ROLE_STATIC)
+            if args.get('string'):
+                doNotSpeak.append("ROLE_CONTENT_SUGGESTION")
             if args.get('formatType', 'unfocused') != 'basicWhereAmI':
                 doNotSpeak.append(pyatspi.ROLE_LIST_ITEM)
                 doNotSpeak.append(pyatspi.ROLE_LIST)
@@ -363,22 +549,24 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
             if self._script.utilities.isMenuInCollapsedSelectElement(obj):
                 doNotSpeak.append(pyatspi.ROLE_MENU)
 
-        if obj.getState().contains(pyatspi.STATE_EDITABLE):
-            lastKey, mods = self._script.utilities.lastKeyAndModifiers()
+        lastKey, mods = self._script.utilities.lastKeyAndModifiers()
+        isEditable = obj.getState().contains(pyatspi.STATE_EDITABLE)
+
+        if isEditable and not self._script.utilities.isContentEditableWithEmbeddedObjects(obj):
             if ((lastKey in ["Down", "Right"] and not mods) or self._script.inSayAll()) and start:
                 return []
             if lastKey in ["Up", "Left"] and not mods:
                 text = self._script.utilities.queryNonEmptyText(obj)
                 if text and end not in [None, text.characterCount]:
                     return []
-            if role in [pyatspi.ROLE_ENTRY, pyatspi.ROLE_PASSWORD_TEXT, pyatspi.ROLE_SPIN_BUTTON]:
+            if role not in doNotSpeak:
                 result.append(self.getLocalizedRoleName(obj, **args))
-            elif obj.parent and not obj.parent.getState().contains(pyatspi.STATE_EDITABLE):
-                if lastKey not in ["Home", "End", "Up", "Down", "Left", "Right", "Page_Up", "Page_Down"]:
-                    result.append(object_properties.ROLE_EDITABLE_CONTENT)
-            elif role not in doNotSpeak:
-                result.append(self.getLocalizedRoleName(obj, **args))
-            if result:
+                result.extend(acss)
+
+        elif isEditable and self._script.utilities.isDocument(obj):
+            if obj.parent and not obj.parent.getState().contains(pyatspi.STATE_EDITABLE) \
+               and lastKey not in ["Home", "End", "Up", "Down", "Left", "Right", "Page_Up", "Page_Down"]:
+                result.append(object_properties.ROLE_EDITABLE_CONTENT)
                 result.extend(acss)
 
         elif role == pyatspi.ROLE_HEADING:
@@ -405,11 +593,6 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                     result.append(self.getLocalizedRoleName(obj, **args))
                     result.extend(acss)
 
-        elif role == pyatspi.ROLE_COMMENT:
-            if index == 0:
-                result.append(self.getLocalizedRoleName(obj, **args))
-                result.extend(acss)
-
         elif role not in doNotSpeak and args.get('priorObj') != obj:
             result.append(self.getLocalizedRoleName(obj, **args))
             result.extend(acss)
@@ -508,6 +691,10 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         if self._script.utilities.isEditableComboBox(obj):
             return []
 
+        if args.get('formatType') not in ['basicWhereAmI', 'detailedWhereAmI']:
+            if args.get('priorObj') == obj:
+                return []
+
         position = self._script.utilities.getPositionInSet(obj)
         total = self._script.utilities.getSetSize(obj)
         if position is None or total is None:
@@ -527,6 +714,12 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         result.extend(self.voice(speech_generator.SYSTEM))
         return result
 
+    def _generateUnselectedCell(self, obj, **args):
+        if not self._script.inFocusMode():
+            return []
+
+        return super()._generateUnselectedCell(obj, **args)
+
     def _generateRealTableCell(self, obj, **args):
         result = super()._generateRealTableCell(obj, **args)
         if not self._script.inFocusMode():
@@ -550,7 +743,7 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         return result
 
     def _generateTableCellRow(self, obj, **args):
-        if not self._script.inFocusMode():
+        if not self._script.utilities.inDocumentContent(obj):
             return super()._generateTableCellRow(obj, **args)
 
         if not self._script.utilities.shouldReadFullRow(obj):
@@ -563,6 +756,12 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
 
         return super()._generateTableCellRow(obj, **args)
 
+    def _generateRowHeader(self, obj, **args):
+        if self._script.utilities.lastInputEventWasLineNav():
+            return []
+
+        return super()._generateRowHeader(obj)
+
     def generateSpeech(self, obj, **args):
         if not self._script.utilities.inDocumentContent(obj):
             msg = "WEB: %s is not in document content. Calling default speech generator." % obj
@@ -577,26 +776,23 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
             oldRole = self._overrideRole('default', args)
         elif self._script.utilities.isLink(obj):
             oldRole = self._overrideRole(pyatspi.ROLE_LINK, args)
+        elif self._script.utilities.isCustomImage(obj):
+            oldRole = self._overrideRole(pyatspi.ROLE_IMAGE, args)
         elif self._script.utilities.treatAsDiv(obj, offset=args.get('startOffset')):
             oldRole = self._overrideRole(pyatspi.ROLE_SECTION, args)
         else:
             oldRole = self._overrideRole(self._getAlternativeRole(obj, **args), args)
 
         if not 'priorObj' in args:
-            args['priorObj'] = self._script.utilities.getPriorContext()[0]
-
-        if self._script.utilities.isLabellingContents(obj):
-            result = list(filter(lambda x: x, self.generateContext(obj, **args)))
+            document = self._script.utilities.getTopLevelDocumentForObject(obj)
+            args['priorObj'] = self._script.utilities.getPriorContext(document)[0]
 
         if not result:
             result = list(filter(lambda x: x, super().generateSpeech(obj, **args)))
 
         self._restoreRole(oldRole, args)
-        msg = "WEB: Speech generation for document object %s complete:" % obj
-        debug.println(debug.LEVEL_INFO, msg)
-        for element in result:
-            debug.println(debug.LEVEL_ALL, "           %s" % element)
-
+        msg = "WEB: Speech generation for document object %s complete." % obj
+        debug.println(debug.LEVEL_INFO, msg, True)
         return result
 
     def generateContents(self, contents, **args):
@@ -623,7 +819,8 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                 args['priorObj'] = obj
 
         if not result:
-            if self._script.inSayAll():
+            if self._script.inSayAll(treatInterruptedAsIn=False) \
+               or not _settingsManager.getSetting('speakBlankLines'):
                 string = ""
             else:
                 string = messages.BLANK
