@@ -34,9 +34,11 @@ import os
 from gi.repository import Gio, GLib
 
 from . import debug
+from . import orca_i18n
 from . import script_manager
 from . import settings
 from . import pronunciation_dict
+from .acss import ACSS
 from .keybindings import KeyBinding
 
 try:
@@ -71,7 +73,7 @@ class SettingsManager(object):
         backend='json'
         """
 
-        debug.println(debug.LEVEL_FINEST, 'INFO: Initializing settings manager')
+        debug.println(debug.LEVEL_INFO, 'SETTINGS MANAGER: Initializing', True)
 
         self.backendModule = None
         self._backend = None
@@ -82,10 +84,7 @@ class SettingsManager(object):
         # Dictionaries for store the default values
         # The keys and values are defined at orca.settings
         #
-        ## self.defaultGeneral contain some constants names as values
         self.defaultGeneral = {}
-        ## self.defaultGeneralValues contain the actual values, no constants
-        self.defaultGeneralValues = {}
         self.defaultPronunciations = {}
         self.defaultKeybindings = {}
 
@@ -103,6 +102,11 @@ class SettingsManager(object):
         self.pronunciations = {}
         self.keybindings = {}
 
+        self._activeApp = ""
+        self._appGeneral = {}
+        self._appPronunciations = {}
+        self._appKeybindings = {}
+
         if not self._loadBackend():
             raise Exception('SettingsManager._loadBackend failed.')
 
@@ -111,12 +115,11 @@ class SettingsManager(object):
 
         # For handling the currently-"classic" application settings
         self.settingsPackages = ["app-settings"]
-        self._knownAppSettings = {}
 
-        debug.println(debug.LEVEL_FINEST, 'INFO: Settings manager initialized')
+        debug.println(debug.LEVEL_INFO, 'SETTINGS MANAGER: Initialized', True)
 
     def activate(self, prefsDir=None, customSettings={}):
-        debug.println(debug.LEVEL_FINEST, 'INFO: Activating settings manager')
+        debug.println(debug.LEVEL_INFO, 'SETTINGS MANAGER: Activating', True)
 
         self.customizedSettings.update(customSettings)
         self._prefsDir = prefsDir \
@@ -127,8 +130,7 @@ class SettingsManager(object):
         self._setDefaultGeneral()
         self._setDefaultPronunciations()
         self._setDefaultKeybindings()
-        self.defaultGeneralValues = getRealValues(self.defaultGeneral)
-        self.general = self.defaultGeneralValues.copy()
+        self.general = self.defaultGeneral.copy()
         if not self.isFirstStart():
             self.general.update(self._backend.getGeneral())
         self.pronunciations = self.defaultPronunciations.copy()
@@ -140,11 +142,17 @@ class SettingsManager(object):
         #
         self._createDefaults()
 
-        debug.println(debug.LEVEL_FINEST, 'INFO: Settings manager activated')
+        debug.println(debug.LEVEL_INFO, 'SETTINGS MANAGER: Activated', True)
 
         # Set the active profile and load its stored settings
+        msg = 'SETTINGS MANAGER: Current profile is %s' % self.profile
+        debug.println(debug.LEVEL_INFO, msg, True)
+
         if self.profile is None:
             self.profile = self.general.get('startingProfile')[1]
+            msg = 'SETTINGS MANAGER: Current profile is now %s' % self.profile
+            debug.println(debug.LEVEL_INFO, msg, True)
+
         self.setProfile(self.profile)
 
     def _loadBackend(self):
@@ -179,13 +187,11 @@ class SettingsManager(object):
         if not os.path.exists(initFile):
             os.close(os.open(initFile, os.O_CREAT, 0o700))
 
-        # Set up $XDG_DATA_HOME/orca/app-settings as a Python package.
-        #
         orcaSettingsDir = os.path.join(orcaDir, "app-settings")
         _createDir(orcaSettingsDir)
-        initFile = os.path.join(orcaSettingsDir, "__init__.py")
-        if not os.path.exists(initFile):
-            os.close(os.open(initFile, os.O_CREAT, 0o700))
+
+        orcaSoundsDir = os.path.join(orcaDir, "sounds")
+        _createDir(orcaSoundsDir)
 
         # Set up $XDG_DATA_HOME/orca/orca-customizations.py empty file and
         # define orcaDir as a Python package.
@@ -216,7 +222,7 @@ class SettingsManager(object):
         self.defaultGeneral = {}
         for key in settings.userCustomizableSettings:
             value = self.customizedSettings.get(key)
-            if value == None:
+            if value is None:
                 try:
                     value = getattr(settings, key)
                 except:
@@ -228,12 +234,12 @@ class SettingsManager(object):
             return self.customizedSettings
 
         originalSettings = {}
-        for key, value in list(settings.__dict__.items()):
+        for key, value in settings.__dict__.items():
             originalSettings[key] = value
 
         self._customizationCompleted = self._loadUserCustomizations()
 
-        for key, value in list(originalSettings.items()):
+        for key, value in originalSettings.items():
             customValue = settings.__dict__.get(key)
             if value != customValue:
                 self.customizedSettings[key] = customValue
@@ -246,7 +252,7 @@ class SettingsManager(object):
         success = False
         pathList = [self._prefsDir]
         try:
-            msg = "Attempt to load orca-customizations "
+            msg = "SETTINGS MANAGER: Attempt to load orca-customizations "
             (fileHnd, moduleName, desc) = \
                 imp.find_module("orca-customizations", pathList)
             msg += "from %s " % moduleName
@@ -260,7 +266,7 @@ class SettingsManager(object):
             msg += "succeeded."
             fileHnd.close()
             success = True
-        debug.println(debug.LEVEL_ALL, msg)
+        debug.println(debug.LEVEL_ALL, msg, True)
         return success
 
     def getPrefsDir(self):
@@ -270,48 +276,53 @@ class SettingsManager(object):
         self._setSettingsRuntime({settingName:settingValue})
 
     def getSetting(self, settingName):
-        return getattr(settings, settingName)
+        return getattr(settings, settingName, None)
 
-    def _getGeneral(self, profile=None):
-        """Get from the active backend the general settings for
-        the current profile"""
-        if profile is None:
-            profile = self.profile
-        self.general = self._backend.getGeneral(profile)
-
-    def _getPronunciations(self, profile=None):
-        """Get from the active backend the pronunciations settings for
-        the current profile"""
-        if profile is None:
-            profile = self.profile
-        self.pronunciations = self._backend.getPronunciations(profile)
-
-    def _getKeybindings(self, profile=None):
-        """Get from the active backend the keybindings settings for
-        the current profile"""
-        if profile is None:
-            profile = self.profile
-        self.keybindings = self._backend.getKeybindings(profile)
+    def getVoiceLocale(self, voice='default'):
+        voices = self.getSetting('voices')
+        v = ACSS(voices.get(voice, {}))
+        lang = v.getLocale()
+        dialect = v.getDialect()
+        if dialect and len(str(dialect)) == 2:
+            lang = "%s_%s" % (lang, dialect.upper())
+        return lang
 
     def _loadProfileSettings(self, profile=None):
         """Get from the active backend all the settings for the current
         profile and store them in the object's attributes.
         A profile can be passed as a parameter. This could be useful for
         change from one profile to another."""
+
+        msg = 'SETTINGS MANAGER: Loading settings for %s profile' % profile
+        debug.println(debug.LEVEL_INFO, msg, True)
+
         if profile is None:
             profile = self.profile
         self.profileGeneral = self.getGeneralSettings(profile) or {}
         self.profilePronunciations = self.getPronunciations(profile) or {}
         self.profileKeybindings = self.getKeybindings(profile) or {}
 
+        msg = 'SETTINGS MANAGER: Settings for %s profile loaded' % profile
+        debug.println(debug.LEVEL_INFO, msg, True)
+
     def _mergeSettings(self):
         """Update the changed values on the profile settings
         over the current and active settings"""
-        profileGeneral = getRealValues(self.profileGeneral) or {}
-        self.general.update(profileGeneral)
+
+        msg = 'SETTINGS MANAGER: Merging settings.'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
+        self.profileGeneral.update(self._appGeneral)
+        self.profilePronunciations.update(self._appPronunciations)
+        self.profileKeybindings.update(self._appKeybindings)
+
+        self.general.update(self.profileGeneral)
         self.pronunciations.update(self.profilePronunciations)
         self.keybindings.update(self.profileKeybindings)
 
+        msg = 'SETTINGS MANAGER: Settings merged.'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
     def _enableAccessibility(self):
         """Enables the GNOME accessibility flag.  Users need to log out and
         then back in for this to take effect.
@@ -327,60 +338,106 @@ class SettingsManager(object):
         return not alreadyEnabled
 
     def isAccessibilityEnabled(self):
-        if not _proxy:
-            return False
+        msg = 'SETTINGS MANAGER: Checking if accessibility is enabled.'
+        debug.println(debug.LEVEL_INFO, msg, True)
 
-        return _proxy.Get('(ss)', 'org.a11y.Status', 'IsEnabled')
+        msg = 'SETTINGS MANAGER: Accessibility enabled: '
+        if not _proxy:
+            rv = False
+            msg += 'Error (no proxy)'
+        else:
+            rv = _proxy.Get('(ss)', 'org.a11y.Status', 'IsEnabled')
+            msg += str(rv)
+
+        debug.println(debug.LEVEL_INFO, msg, True)
+        return rv
 
     def setAccessibility(self, enable):
+        msg = 'SETTINGS MANAGER: Attempting to set accessibility to %s.' % enable
+        debug.println(debug.LEVEL_INFO, msg, True)
+
         if not _proxy:
+            msg = 'SETTINGS MANAGER: Error (no proxy)'
+            debug.println(debug.LEVEL_INFO, msg, True)
             return False
 
         vEnable = GLib.Variant('b', enable)
-        _proxy.Set('(ssv)', 'org.a11y.Status', 'IsEnabled', vEnable)            
+        _proxy.Set('(ssv)', 'org.a11y.Status', 'IsEnabled', vEnable)
+
+        msg = 'SETTINGS MANAGER: Finished setting accessibility to %s.' % enable
+        debug.println(debug.LEVEL_INFO, msg, True)
 
     def isScreenReaderServiceEnabled(self):
         """Returns True if the screen reader service is enabled. Note that
         this does not necessarily mean that Orca (or any other screen reader)
         is running at the moment."""
 
-        if not _proxy:
-            return False
+        msg = 'SETTINGS MANAGER: Is screen reader service enabled? '
 
-        return _proxy.Get('(ss)', 'org.a11y.Status', 'ScreenReaderEnabled')
+        if not _proxy:
+            rv = False
+            msg += 'Error (no proxy)'
+        else:
+            rv = _proxy.Get('(ss)', 'org.a11y.Status', 'ScreenReaderEnabled')
+            msg += str(rv)
+
+        debug.println(debug.LEVEL_INFO, msg, True)
+        return rv
 
     def setStartingProfile(self, profile=None):
         if profile is None:
             profile = settings.profile
         self._backend._setProfileKey('startingProfile', profile)
 
-    def setProfile(self, profile='default'):
+    def getProfile(self):
+        return self.profile
+
+    def setProfile(self, profile='default', updateLocale=False):
         """Set a specific profile as the active one.
         Also the settings from that profile will be loading
         and updated the current settings with them."""
+
+        msg = 'SETTINGS MANAGER: Setting profile to: %s' % profile
+        debug.println(debug.LEVEL_INFO, msg, True)
+
+        oldVoiceLocale = self.getVoiceLocale('default')
+
         self.profile = profile
         self._loadProfileSettings(profile)
         self._mergeSettings()
         self._setSettingsRuntime(self.general)
 
-    def getPreferences(self, profile='default'):
-        general = self.getGeneralSettings(profile)
-        pronunciations = self.getPronunciations(profile)
-        keybindings = self.getKeybindings(profile)
-        return (general, pronunciations, keybindings)
+        if not updateLocale:
+            return
+
+        newVoiceLocale = self.getVoiceLocale('default')
+        if oldVoiceLocale != newVoiceLocale:
+            orca_i18n.setLocaleForNames(newVoiceLocale)
+            orca_i18n.setLocaleForMessages(newVoiceLocale)
+            orca_i18n.setLocaleForGUI(newVoiceLocale)
+
+        msg = 'SETTINGS MANAGER: Profile set to: %s' % profile
+        debug.println(debug.LEVEL_INFO, msg, True)
+
+    def removeProfile(self, profile):
+        self._backend.removeProfile(profile)
 
     def _setSettingsRuntime(self, settingsDict):
-        for key, value in list(settingsDict.items()):
+        msg = 'SETTINGS MANAGER: Setting runtime settings.'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
+        for key, value in settingsDict.items():
             setattr(settings, str(key), value)
         self._getCustomizedSettings()
-        for key, value in list(self.customizedSettings.items()):
+        for key, value in self.customizedSettings.items():
             setattr(settings, str(key), value)
-        self._setPronunciationsRuntime()
 
-    def _setPronunciationsRuntime(self):
+        msg = 'SETTINGS MANAGER: Runtime settings set.'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
+    def _setPronunciationsRuntime(self, pronunciationsDict):
         pronunciation_dict.pronunciation_dict = {}
-        for pron in self.pronunciations:
-            key, value = self.pronunciations[pron]
+        for key, value in pronunciationsDict.values():
             if key and value:
                 pronunciation_dict.setPronunciation(key, value)
 
@@ -388,9 +445,7 @@ class SettingsManager(object):
         """Return the current general settings.
         Those settings comes from updating the default settings
         with the profiles' ones"""
-        generalDict = self._backend.getGeneral(profile)
-        self._setSettingsRuntime(generalDict)
-        return generalDict
+        return self._backend.getGeneral(profile)
 
     def getPronunciations(self, profile='default'):
         """Return the current pronunciations settings.
@@ -407,35 +462,87 @@ class SettingsManager(object):
     def _setProfileGeneral(self, general):
         """Set the changed general settings from the defaults' ones
         as the profile's."""
+
+        msg = 'SETTINGS MANAGER: Setting general settings for profile'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
         self.profileGeneral = {}
 
-        for key, value in list(general.items()):
-            if isinstance(value, unicode):
-                value = value.encode('UTF-8')
-            if key in settings.excludeKeys:
+        for key, value in general.items():
+            if key in ['startingProfile', 'activeProfile']:
                 continue
             elif key == 'profile':
                 self.profileGeneral[key] = value
-            elif value != self.defaultGeneralValues.get(key):
+            elif value != self.defaultGeneral.get(key):
                 self.profileGeneral[key] = value
             elif self.general.get(key) != value:
                 self.profileGeneral[key] = value
 
+        msg = 'SETTINGS MANAGER: General settings for profile set'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
     def _setProfilePronunciations(self, pronunciations):
         """Set the changed pronunciations settings from the defaults' ones
         as the profile's."""
+
+        msg = 'SETTINGS MANAGER: Setting pronunciation settings for profile.'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
         self.profilePronunciations = self.defaultPronunciations.copy()
         self.profilePronunciations.update(pronunciations)
 
+        msg = 'SETTINGS MANAGER: Pronunciation settings for profile set.'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
     def _setProfileKeybindings(self, keybindings):
         """Set the changed keybindings settings from the defaults' ones
         as the profile's."""
+
+        msg = 'SETTINGS MANAGER: Setting keybindings settings for profile.'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
         self.profileKeybindings = self.defaultKeybindings.copy()
         self.profileKeybindings.update(keybindings)
 
-    def saveSettings(self, general, pronunciations, keybindings):
-        """Let the active backend to store the default settings and
-        the profiles' ones."""
+        msg = 'SETTINGS MANAGER: Keybindings settings for profile set.'
+        debug.println(debug.LEVEL_INFO, msg, True)
+
+    def _saveAppSettings(self, appName, general, pronunciations, keybindings):
+        appGeneral = {}
+        profileGeneral = self.getGeneralSettings(self.profile)
+        for key, value in general.items():
+            if value != profileGeneral.get(key):
+                appGeneral[key] = value
+
+        appPronunciations = {}
+        profilePronunciations = self.getPronunciations(self.profile)
+        for key, value in pronunciations.items():
+            if value != profilePronunciations.get(key):
+                appPronunciations[key] = value
+
+        appKeybindings = {}
+        profileKeybindings = self.getKeybindings(self.profile)
+        for key, value in keybindings.items():
+            if value != profileKeybindings.get(key):
+                appKeybindings[key] = value
+
+        self._backend.saveAppSettings(appName,
+                                      self.profile,
+                                      appGeneral,
+                                      appPronunciations,
+                                      appKeybindings)
+
+    def saveSettings(self, script, general, pronunciations, keybindings):
+        """Save the settings provided for the script provided."""
+
+        msg = 'SETTINGS MANAGER: Saving settings for %s (app: %s)' % (script, script.app)
+        debug.println(debug.LEVEL_INFO, msg, True)
+
+        app = script.app
+        if app:
+            self._saveAppSettings(app.name, general, pronunciations, keybindings)
+            return
+
         # Assign current profile
         _profile = general.get('profile', settings.profile)
         currentProfile = _profile[1]
@@ -450,10 +557,17 @@ class SettingsManager(object):
         self._setProfilePronunciations(pronunciations)
         self._setProfileKeybindings(keybindings)
 
+        msg = 'SETTINGS MANAGER: Saving for backend %s' % self._backend
+        debug.println(debug.LEVEL_INFO, msg, True)
+
         self._backend.saveProfileSettings(self.profile,
                                           self.profileGeneral,
                                           self.profilePronunciations,
                                           self.profileKeybindings)
+
+        msg = 'SETTINGS MANAGER: Settings for %s (app: %s) saved' % (script, script.app)
+        debug.println(debug.LEVEL_INFO, msg, True)
+
         return self._enableAccessibility()
 
     def _adjustBindingTupleValues(self, bindingTuple):
@@ -468,8 +582,8 @@ class SettingsManager(object):
         return bindingTuple
 
     def overrideKeyBindings(self, script, scriptKeyBindings):
-        keybindingsSettings = self.getKeybindings(self.profile)
-        for handlerString, bindingTuples in list(keybindingsSettings.items()):
+        keybindingsSettings = self.profileKeybindings
+        for handlerString, bindingTuples in keybindingsSettings.items():
             handler = script.inputEventHandlers.get(handlerString)
             if not handler:
                 continue
@@ -488,7 +602,7 @@ class SettingsManager(object):
         return self._backend.isFirstStart()
 
     def setFirstStart(self, value=False):
-        """Set firstStart. This user-configurable settting is primarily
+        """Set firstStart. This user-configurable setting is primarily
         intended to serve as an indication as to whether or not initial
         configuration is needed."""
         self._backend.setFirstStart(value)
@@ -498,105 +612,48 @@ class SettingsManager(object):
 
         return self._backend.availableProfiles()
 
+    def getAppSetting(self, app, settingName, fallbackOnDefault=True):
+        if not app:
+            return None
+
+        appPrefs = self._backend.getAppSettings(app.name)
+        profiles = appPrefs.get('profiles', {})
+        profilePrefs = profiles.get(self.profile, {})
+        general = profilePrefs.get('general', {})
+        appSetting = general.get(settingName)
+        if appSetting is None and fallbackOnDefault:
+            general = self._backend.getGeneral(self.profile)
+            appSetting = general.get(settingName)
+
+        return appSetting
+
     def loadAppSettings(self, script):
         """Load the users application specific settings for an app.
-        Note that currently the settings manager does not manage
-        application settings in Orca; instead the old/"classic" files
-        are used. This is scheduled to change.
 
         Arguments:
         - script: the current active script.
         """
 
+        if not (script and script.app):
+            return
+
+        for key in self._appPronunciations.keys():
+            self.pronunciations.pop(key)
+
+        prefs = self._backend.getAppSettings(script.app.name)
+        profiles = prefs.get('profiles', {})
+        profilePrefs = profiles.get(self.profile, {})
+
+        self._appGeneral = profilePrefs.get('general', {})
+        self._appKeybindings = profilePrefs.get('keybindings', {})
+        self._appPronunciations = profilePrefs.get('pronunciations', {})
+        self._activeApp = script.app.name
+
         self._loadProfileSettings()
-        script.voices = self.getSetting('voices')
-
-        app = script.app
-        moduleName = _scriptManager.getModuleName(app)
-        if not moduleName:
-            return
-
-        module = None
-        for package in self.settingsPackages:
-            name = '.'.join((package, moduleName))
-            debug.println(debug.LEVEL_FINEST, "Looking for %s.py" % name)
-            try:
-                module = importlib.import_module(name)
-            except ImportError:
-                debug.println(
-                    debug.LEVEL_FINEST, "Could not import %s.py" % name)
-                continue
-            try:
-                imp.reload(module)
-            except:
-                debug.println(debug.LEVEL_FINEST, "Could not load %s.py" % name)
-                module = None
-            else:
-                debug.println(debug.LEVEL_FINEST, "Loaded %s.py" % name)
-                break
-
-        if not module:
-            return
-
-        self._knownAppSettings[name] = module
-        imp.reload(self._knownAppSettings[name])
-
-        appVoices = self.getSetting('voices')
-        for voiceType, voiceDef in list(appVoices.items()):
-            script.voices[voiceType].update(voiceDef)
-
-        keybindings = getattr(module, 'overrideAppKeyBindings', None)
-        if keybindings:
-            script.overrideAppKeyBindings = keybindings
-            script.keyBindings = keybindings(script, script.keyBindings)
-
-        pronunciations = getattr(module, 'overridePronunciations', None)
-        if pronunciations:
-            script.overridePronunciations = pronunciations
-            script.app_pronunciation_dict = \
-                pronunciations(script, script.app_pronunciation_dict)
-
-def getVoiceKey(voice):
-    voicesKeys = getattr(settings, 'voicesKeys')
-    for key in list(voicesKeys.keys()):
-        if voicesKeys[key] == voice:
-            return key
-    return ""
-
-def getValueForKey(prefsDict, key):
-    need2repr = ['brailleEOLIndicator', 'brailleContractionTable',
-                 'brailleRequiredStateString', 'enabledBrailledTextAttributes',
-                 'enabledSpokenTextAttributes', 'speechRequiredStateString',
-                 'speechServerFactory', 'presentDateFormat',
-                 'presentTimeFormat']
-
-    value = None
-    if key in prefsDict:
-        if isinstance(prefsDict[key], str):
-            if key in need2repr:
-                value = "\'%s\'" % prefsDict[key]
-            elif key  == 'voices':
-                key = getVoiceKey(key)
-                value = prefsDict[key]
-            else:
-                try:
-                    value = getattr(settings, prefsDict[key])
-                except:
-                    debug.println(debug.LEVEL_SEVERE,
-                                  "Something went wront with key: " % key)
-                    debug.printStack(debug.LEVEL_FINEST)
-        else:
-            value = prefsDict[key]
-    return value
-
-def getRealValues(prefs):
-    """Get the actual values for any constant stored on a
-    general settings dictionary.
-    prefs is a dictionary with the userCustomizableSettings keys
-    and values."""
-    #for key in prefs.keys():
-    #    prefs[key] = getValueForKey(prefs, key)
-    return prefs
+        self._mergeSettings()
+        self._setSettingsRuntime(self.general)
+        self._setPronunciationsRuntime(self.pronunciations)
+        script.keyBindings = self.overrideKeyBindings(script, script.getKeyBindings())
 
 _manager = SettingsManager()
 
