@@ -248,10 +248,11 @@ class SpeechServer(speechserver.SpeechServer):
 
     def _set_family(self, acss_family):
         lang, dialect = self._get_language_and_dialect(acss_family)
-        self._send_command(self._client.set_language, lang)
-        if dialect:
-            # Try to set precise dialect
-            self._send_command(self._client.set_language, lang + '-' + dialect)
+        if lang:
+            self._send_command(self._client.set_language, lang)
+            if dialect:
+                # Try to set precise dialect
+                self._send_command(self._client.set_language, lang + '-' + dialect)
 
         try:
             # This command is not available with older SD versions.
@@ -260,7 +261,8 @@ class SpeechServer(speechserver.SpeechServer):
             pass
         else:
             name = acss_family.get(speechserver.VoiceFamily.NAME)
-            self._send_command(set_synthesis_voice, name)
+            if name != self._default_voice_name:
+                self._send_command(set_synthesis_voice, name)
 
     def _debug_sd_values(self, prefix=""):
         if debug.debugLevel > debug.LEVEL_INFO:
@@ -371,6 +373,7 @@ class SpeechServer(speechserver.SpeechServer):
         # Note2: we assume that text mangling below leave U+E000 untouched
         last_begin = None
         last_end = None
+        is_numeric = None
         marks_offsets = []
         marks_endoffsets = []
         marked_text = ""
@@ -386,12 +389,33 @@ class SpeechServer(speechserver.SpeechServer):
                 # Word begin
                 marked_text += '\ue000'
                 last_begin = i
+                is_numeric = c.isnumeric()
 
-            if c.isspace() and last_begin != None:
-                # Word end, add a mark
-                marks_offsets.append(last_begin)
-                marks_endoffsets.append(i)
-                last_begin = None
+            elif c.isspace() and last_begin != None:
+                # Word end
+                if is_numeric:
+                    # We had a wholy numeric word, possibly next word is as well.
+                    # Skip to next word
+                    for j in range(i+1, len(text)):
+                        if not text[j].isspace():
+                            break
+                    else:
+                        is_numeric = False
+                    # Check next word
+                    while is_numeric and j < len(text) and not text[j].isspace():
+                        if not text[j].isnumeric():
+                            is_numeric = False
+                        j += 1
+
+                if not is_numeric:
+                    # add a mark
+                    marks_offsets.append(last_begin)
+                    marks_endoffsets.append(i)
+                    last_begin = None
+                    is_numeric = None
+
+            elif is_numeric and not c.isnumeric():
+                is_numeric = False
 
             marked_text += c
 
@@ -540,7 +564,7 @@ class SpeechServer(speechserver.SpeechServer):
         # set according to the current locale.
         from locale import getlocale, LC_MESSAGES
         locale = getlocale(LC_MESSAGES)[0]
-        if locale is None or locale == 'C':
+        if locale is None or not '_' in locale:
             locale_language = None
         else:
             locale_lang, locale_dialect = locale.split('_')
@@ -586,6 +610,12 @@ class SpeechServer(speechserver.SpeechServer):
         return families
 
     def speak(self, text=None, acss=None, interrupt=True):
+        # In order to re-enable this, a potentially non-trivial amount of work
+        # will be needed to ensure multiple utterances sent to speech.speak
+        # do not result in the intial utterances getting cut off before they
+        # can be heard by the user. Anyone needing to interrupt speech can
+        # do so via speech.stop -- or better yet, by using the default script
+        # method's presentationInterrupt.
         #if interrupt:
         #    self._cancel()
 
@@ -600,6 +630,12 @@ class SpeechServer(speechserver.SpeechServer):
             self._speak(text, acss)
 
     def speakUtterances(self, utteranceList, acss=None, interrupt=True):
+        # In order to re-enable this, a potentially non-trivial amount of work
+        # will be needed to ensure multiple utterances sent to speech.speak
+        # do not result in the intial utterances getting cut off before they
+        # can be heard by the user. Anyone needing to interrupt speech can
+        # do so via speech.stop -- or better yet, by using the default script
+        # method's presentationInterrupt.
         #if interrupt:
         #    self._cancel()
         for utterance in utteranceList:
